package pl.net.bluesoft.rnd.processtool.plugins.osgi;

import static pl.net.bluesoft.rnd.processtool.plugins.ProcessToolRegistry.Util.getRegistry;
import static pl.net.bluesoft.rnd.processtool.plugins.osgi.OSGiBundleHelper.*;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.Collection;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.osgi.framework.Bundle;

import pl.net.bluesoft.rnd.processtool.ProcessToolContext;
import pl.net.bluesoft.rnd.processtool.ProcessToolContextCallback;
import pl.net.bluesoft.rnd.processtool.di.ObjectFactory;
import pl.net.bluesoft.rnd.processtool.plugins.IBundleResourceProvider;
import pl.net.bluesoft.rnd.processtool.plugins.ProcessToolRegistry;
import pl.net.bluesoft.rnd.processtool.plugins.deployment.ProcessDeployer;
import pl.net.bluesoft.rnd.processtool.plugins.osgi.beans.ScriptFileNameBean;
import pl.net.bluesoft.rnd.processtool.roles.IUserRolesManager;
import pl.net.bluesoft.rnd.processtool.steps.ProcessToolProcessStep;
import pl.net.bluesoft.rnd.processtool.web.controller.IOsgiWebController;
import pl.net.bluesoft.rnd.processtool.web.controller.OsgiController;
import pl.net.bluesoft.rnd.processtool.web.domain.IWidgetScriptProvider;
import pl.net.bluesoft.rnd.processtool.web.widgets.impl.FileWidgetJavaScriptProvider;
import pl.net.bluesoft.rnd.processtool.ui.widgets.ProcessHtmlWidget;
import pl.net.bluesoft.rnd.util.i18n.impl.PropertiesBasedI18NProvider;
import pl.net.bluesoft.rnd.util.i18n.impl.PropertyLoader;

import com.thoughtworks.xstream.XStream;

/**
 * User: POlszewski
 * Date: 2012-11-27
 * Time: 16:17
 */
public class BundleInstallationHandler {
	private static final String SEPARATOR = "/";

	private ProcessToolRegistry registry;
	private ErrorMonitor errorMonitor;
	private Logger logger;

	public BundleInstallationHandler(ErrorMonitor errorMonitor, Logger logger) {
		this.errorMonitor = errorMonitor;
		this.logger = logger;
	}

	public synchronized void processBundleExtensions(Bundle bundle, int eventType) throws ClassNotFoundException {
		if (registry.getProcessToolContextFactory() == null) {
			logger.severe("No default process tool context registered! - skipping process tool context-based processing of this OSGI bundle");
			return;
		}

		OSGiBundleHelper bundleHelper = new OSGiBundleHelper(bundle);
		
		if (bundleHelper.hasHeaderValues(VIEW)) {
			handleView(eventType, bundleHelper, registry);
		}
		
		if (bundleHelper.hasHeaderValues(SCRIPT)) {
			handleScript(eventType, bundleHelper, registry);
		}

        if (bundleHelper.hasHeaderValues(CONTROLLER)) {
            handleController(eventType, bundleHelper, registry);
        }

		if (bundleHelper.hasHeaderValues(MODEL_ENHANCEMENT)) {
			handleModelEnhancement(eventType, bundleHelper, registry);
		}

		if (bundleHelper.hasHeaderValues(WIDGET_ENHANCEMENT)) {
			handleWidgetEnhancement(eventType, bundleHelper, registry);
		}

		if (bundleHelper.hasHeaderValues(BUTTON_ENHANCEMENT)) {
			handleButtonEnhancement(eventType, bundleHelper, registry);
		}

		if (bundleHelper.hasHeaderValues(STEP_ENHANCEMENT)) {
			handleStepEnhancement(eventType, bundleHelper, registry);
		}

		if (bundleHelper.hasHeaderValues(I18N_PROPERTY)) {
			handleMessageSources(eventType, bundleHelper, registry);
		}

		if (bundleHelper.hasHeaderValues(PROCESS_DEPLOYMENT)) {
			handleProcessRoles(eventType, bundleHelper);
			handleProcessDeployment(eventType, bundleHelper, registry);
		}

		if (bundleHelper.hasHeaderValues(GLOBAL_DICTIONARY)) {
			handleGlobalDictionaries(eventType, bundleHelper, registry);
		}

		if (bundleHelper.hasHeaderValues(RESOURCES)) {
			handleBundleResources(eventType, bundleHelper, registry);
		}
	}

	private void handleScript(int eventType, OSGiBundleHelper bundleHelper,ProcessToolRegistry toolRegistry) 
	{
		Bundle bundle = bundleHelper.getBundle();
		String[] javaScriptFiles = bundleHelper.getHeaderValues(SCRIPT);

		
		for (String javaScriptFileName : javaScriptFiles) 
		{
			try
			{
				ScriptFileNameBean scriptFileNameBean = new ScriptFileNameBean(javaScriptFileName);
				IWidgetScriptProvider scriptProvider;
				
				if(scriptFileNameBean.getProviderClass().equals(FileWidgetJavaScriptProvider.class.getName()))
				{
					scriptProvider = 
							new FileWidgetJavaScriptProvider(
									scriptFileNameBean.getFileName(), 
									bundle.getResource(scriptFileNameBean.getFileName()));
				}
				else
				{
					scriptProvider = (IWidgetScriptProvider)bundle
							.loadClass(scriptFileNameBean.getProviderClass())
							.getConstructor(String.class, URL.class)
							.newInstance(scriptFileNameBean.getFileName());
				}
				
				
				
				if (eventType == Bundle.ACTIVE) 
				{
					toolRegistry.registerJavaScript(scriptFileNameBean.getFileName(), scriptProvider);
				}
				else {
					toolRegistry.unregisterJavaScript(scriptFileNameBean.getFileName());
				}
			}
			catch(Throwable e)
			{
				logger.log(Level.SEVERE, e.getMessage(), e);
				forwardErrorInfoToMonitor(bundle.getSymbolicName(), e);
			}
		}
		
	}

	private void handleView(int eventType, OSGiBundleHelper bundleHelper,ProcessToolRegistry toolRegistry) 
	{
		Bundle bundle = bundleHelper.getBundle();
		String[] widgetClasses = bundleHelper.getHeaderValues(VIEW);

		
		for (String widgetClass : widgetClasses)
		{
			try
			{
                ProcessHtmlWidget htmlWidget = (ProcessHtmlWidget)bundle
                        .loadClass(widgetClass)
                        .getConstructor(IBundleResourceProvider.class)
                        .newInstance(bundleHelper);

				if (eventType == Bundle.ACTIVE) 
				{
					toolRegistry.registerHtmlView(htmlWidget.getWidgetName(), htmlWidget);
				}
				else 
				{
					toolRegistry.unregisterHtmlView(htmlWidget.getWidgetName());
				}
			}
			catch(Throwable e)
			{
				logger.log(Level.SEVERE, e.getMessage(), e);
				forwardErrorInfoToMonitor(bundle.getSymbolicName(), e);
			}
		}
		
	}

    private void handleController(int eventType, OSGiBundleHelper bundleHelper,ProcessToolRegistry toolRegistry)
    {
        Bundle bundle = bundleHelper.getBundle();
        String[] classes = bundleHelper.getHeaderValues(CONTROLLER);

        for (String cls : classes)
        {
            try
            {
                Class<? extends IOsgiWebController> controllerClass =
                        (Class<? extends IOsgiWebController>)bundleHelper.getBundle().loadClass(cls);
                OsgiController controllerAnnotation = controllerClass.getAnnotation(OsgiController.class);

                String controllerName = controllerAnnotation.name();
                if (eventType == Bundle.ACTIVE)
                {
                    IOsgiWebController controller = controllerClass.newInstance();
                    toolRegistry.registerWebController(controllerName, controller);
                }
                else
                {
                    toolRegistry.unregisterWebController(controllerName);
                }
            }
            catch (Throwable e)
            {
                logger.log(Level.SEVERE, e.getMessage(), e);
                forwardErrorInfoToMonitor(bundle.getSymbolicName(), e);
            }
        }

    }

	private void handleMessageSources(int eventType, OSGiBundleHelper bundleHelper, ProcessToolRegistry toolRegistry) {
		final Bundle bundle = bundleHelper.getBundle();
		String[] properties = bundleHelper.getHeaderValues(I18N_PROPERTY);
		for (String propertyFileName : properties) {
			String providerId = bundle.getBundleId() + File.separator + propertyFileName;
			if (eventType == Bundle.ACTIVE) {
				toolRegistry.registerI18NProvider(new PropertiesBasedI18NProvider(new PropertyLoader() {
					@Override
					public InputStream loadProperty(String path) throws IOException {
						return getBundleResourceStream(bundle, path);
					}
				}, propertyFileName), providerId);
			}
			else {
				toolRegistry.unregisterI18NProvider(providerId);
			}
		}
	}

	private void handleStepEnhancement(int eventType, OSGiBundleHelper bundleHelper, ProcessToolRegistry toolRegistry) throws ClassNotFoundException {
		Bundle bundle = bundleHelper.getBundle();
		String[] classes = bundleHelper.getHeaderValues(STEP_ENHANCEMENT);
		for (String cls : classes) {
			if (eventType == Bundle.ACTIVE) {
				toolRegistry.registerStep((Class<? extends ProcessToolProcessStep>) bundle.loadClass(cls));
			}
			else {
				toolRegistry.unregisterStep((Class<? extends ProcessToolProcessStep>) bundle.loadClass(cls));
			}
		}
	}

	private void handleButtonEnhancement(int eventType, OSGiBundleHelper bundleHelper, ProcessToolRegistry toolRegistry) throws ClassNotFoundException {
		Bundle bundle = bundleHelper.getBundle();
		String[] classes = bundleHelper.getHeaderValues(BUTTON_ENHANCEMENT);
		for (String cls : classes) {
			if (eventType == Bundle.ACTIVE) {
				toolRegistry.registerButton(bundle.loadClass(cls));
			}
			else {
				toolRegistry.unregisterButton(bundle.loadClass(cls));
			}
		}
	}

	private void handleWidgetEnhancement(int eventType, OSGiBundleHelper bundleHelper, ProcessToolRegistry toolRegistry) throws ClassNotFoundException {
		Bundle bundle = bundleHelper.getBundle();
		String[] classes = bundleHelper.getHeaderValues(WIDGET_ENHANCEMENT);
		for (String cls : classes) {
			if (eventType == Bundle.ACTIVE) {
				toolRegistry.registerWidget(bundle.loadClass(cls));
			}
			else {
				toolRegistry.unregisterWidget(bundle.loadClass(cls));
			}
		}
	}

	private void handleModelEnhancement(int eventType, OSGiBundleHelper bundleHelper, ProcessToolRegistry toolRegistry) throws ClassNotFoundException {
		Bundle bundle = bundleHelper.getBundle();
		String[] classes = bundleHelper.getHeaderValues(MODEL_ENHANCEMENT);
		Collection<Class> classSet = new HashSet<Class>();
		for (String cls : classes) {
			classSet.add(bundle.loadClass(cls));
		}
		if (!classSet.isEmpty()) {
			Class<?>[] extensions = classSet.toArray(new Class<?>[classSet.size()]);
			boolean needUpdate = eventType == Bundle.ACTIVE
					? toolRegistry.registerModelExtension(extensions)
					: toolRegistry.unregisterModelExtension(extensions);
			if (needUpdate) {
				logger.fine("Rebuilding Hibernate session factory...");
				try {
					toolRegistry.commitModelExtensions();
				}
				catch (Exception e) {
					logger.severe("Encountered problem while updating Hibernate mappings");
					logger.log(Level.SEVERE, e.getMessage(), e);
					toolRegistry.unregisterModelExtension(extensions);
				}
			}
			else {
				logger.warning("Skipping Hibernate session factory rebuild. Classes already processed.");
			}
		}
	}

	private void handleProcessDeployment(int eventType, final OSGiBundleHelper bundleHelper, final ProcessToolRegistry toolRegistry) {
		final Bundle bundle = bundleHelper.getBundle();
		String[] properties = bundleHelper.getHeaderValues(PROCESS_DEPLOYMENT);
		for (final String processPackage : properties) {
			final String providerId = bundle.getBundleId() + SEPARATOR + processPackage.replace(".", SEPARATOR) + "/messages";
			if (eventType == Bundle.ACTIVE) {
				final String basePath = SEPARATOR + processPackage.replace(".", SEPARATOR) + SEPARATOR;

				toolRegistry.withExistingOrNewContext(new ProcessToolContextCallback() {
					@Override
					public void withContext(ProcessToolContext ctx)
					{
						try {
							/* Initialize process deployer */
							ProcessDeployer processDeployer = new ProcessDeployer(ctx);

							processDeployer.deployOrUpdateProcessDefinition(
									bundleHelper.getBundleResourceStream(basePath + "processdefinition." +
											toolRegistry.getBpmDefinitionLanguage()),
											bundleHelper.getBundleResourceStream(basePath + "processtool-config.xml"),
									bundleHelper.getBundleResourceStream(basePath + "queues-config.xml"),
									bundleHelper.getBundleResourceStream(basePath + "processdefinition.png"),
									bundleHelper.getBundleResourceStream(basePath + "processdefinition-logo.png"));

							getRegistry().registerI18NProvider(new PropertiesBasedI18NProvider(new PropertyLoader() {
								@Override
								public InputStream loadProperty(String path) throws IOException {
									return getBundleResourceStream(bundle, path);
								}
							}, "/" + processPackage.replace(".", SEPARATOR) + "/messages"), providerId);
						}
						catch (Exception e) {
							logger.log(Level.SEVERE, e.getMessage(), e);
							forwardErrorInfoToMonitor(bundle.getSymbolicName(), e);
						}
					}
				});
			}
			else { // ignore
				toolRegistry.unregisterI18NProvider(providerId);
			}
		}
	}

	private void handleProcessRoles(int eventType, OSGiBundleHelper bundleHelper) {
		if (eventType != Bundle.ACTIVE) {
			return;
		}

		Bundle bundle = bundleHelper.getBundle();

		if (bundleHelper.hasHeaderValues(ROLE_FILES)) {
			String[] files = bundleHelper.getHeaderValues(ROLE_FILES);
			for (String file : files) {
				try {
					InputStream input = bundleHelper.getBundleResourceStream(file);
					Collection<ProcessRoleConfig> roles = getRoles(input);
					createRoles(roles);
				}
				catch (Exception e) {
					logger.log(Level.SEVERE, e.getMessage(), e);
					forwardErrorInfoToMonitor(bundle.getSymbolicName(), e);
				}
			}
		}

		if (bundleHelper.hasHeaderValues(PROCESS_DEPLOYMENT)) {
			String[] properties = bundleHelper.getHeaderValues(PROCESS_DEPLOYMENT);
			for (String processPackage : properties) {
				String basePath = SEPARATOR + processPackage.replace(".", SEPARATOR) + SEPARATOR;
				try {
					InputStream input = bundleHelper.getBundleResourceStream(basePath + "roles-config.xml");
					Collection<ProcessRoleConfig> roles = getRoles(input);
					createRoles(roles);
				}
				catch (Exception e) {
					logger.log(Level.SEVERE, e.getMessage(), e);
					forwardErrorInfoToMonitor(bundle.getSymbolicName(), e);
				}
			}
		}
	}


	private Collection<ProcessRoleConfig> getRoles(InputStream input) {
		if (input == null) {
			return null;
		}
		XStream xstream = new XStream();
		xstream.aliasPackage("config", ProcessRoleConfig.class.getPackage().getName());
		xstream.useAttributeFor(String.class);
		xstream.useAttributeFor(Boolean.class);
		xstream.useAttributeFor(Integer.class);
		return (Collection<ProcessRoleConfig>) xstream.fromXML(input);
	}

	private void createRoles(Collection<ProcessRoleConfig> roles) {
		if (roles != null) {
			for (ProcessRoleConfig role : roles) {
				try
				{
					IUserRolesManager userRolesManager = ObjectFactory.create(IUserRolesManager.class);
					
					/* Check if role exist. If not, create it */
					if(!userRolesManager.isRoleExist(role.getName()))
					{
						userRolesManager.createRole(role.getName(), role.getDescription());
						logger.log(Level.INFO, "Created role " + role.getName());
					}

				} 
				catch (RuntimeException e) {
					forwardErrorInfoToMonitor("adding role " + role.getName(), e);
					throw e;

				}
			}
		}
	}

	private void handleBundleResources(int eventType, OSGiBundleHelper bundleHelper, ProcessToolRegistry toolRegistry) {
		Bundle bundle = bundleHelper.getBundle();
		String[] resources = bundleHelper.getHeaderValues(RESOURCES);
		for (String pack : resources) {
			if (eventType == Bundle.ACTIVE)
			{
				String basePath = SEPARATOR + pack.replace(".", SEPARATOR);
				if (!basePath.endsWith(SEPARATOR)) {
					basePath += SEPARATOR;
				}
				Enumeration<URL> urls = bundle.findEntries(basePath, null, true);
				while (urls.hasMoreElements()) {
					String path = urls.nextElement().getPath();
					toolRegistry.registerResource(bundle.getSymbolicName(), path);
				}
			}
			else {
				toolRegistry.removeRegisteredResources(bundle.getSymbolicName());
			}
		}
	}

	private void handleGlobalDictionaries(int eventType, OSGiBundleHelper bundleHelper, ProcessToolRegistry toolRegistry) {
		String[] properties = bundleHelper.getHeaderValues(GLOBAL_DICTIONARY);
		if (eventType == Bundle.ACTIVE) {
			for (String pack : properties) {
				try {
					String basePath = SEPARATOR + pack.replace(".", SEPARATOR) + SEPARATOR;
					InputStream is = bundleHelper.getBundleResourceStream(basePath + "global-dictionaries.xml");
					if (is != null) {
						toolRegistry.registerGlobalDictionaries(is);
					}
					else {
						logger.log(Level.SEVERE, "No global dictionary stream found in package: " + pack);
					}
				}
				catch (Exception e) {
					logger.log(Level.SEVERE, e.getMessage(), e);
					forwardErrorInfoToMonitor(bundleHelper.getBundleMetadata().getDescription() + " global-dictionary", e);
				}
			}
		}
	}

	public void setRegistry(ProcessToolRegistry registry) {
		this.registry = registry;
	}

	private void forwardErrorInfoToMonitor(String path, Throwable e) {
		errorMonitor.forwardErrorInfoToMonitor(path, e);
	}
}
